/*
4-digit 7-segment clock with 2x 74hc595 shift registers
This takes a number from SoftwareSerial communication, e.g., 1259 or 100, and displays it on a 4-digit 7 segment display
such as 12:59 or 1:00
When not recieving a new number from the serial communication it keeps displaying that number
It uses persistance of vision as the ATtiny85 sends one digit a a time to the display
(1 to the 4th digit, 2 to the 3rd digit, 5 to the 2nd digit, 9 to the first digit)
If the 10's digit for hour is not a one, then that is just left blank

Notes: Limit to baud rate: 115200 did Not work;57600 & 74880 worked;


Obviously you need someting to send the serial data to the ATtiny85, etc.

The Common Cathode 4-segment 7-digit display I have is:
  1) the Pinout of the 4-digit 7-segment display 6 pins on the top, and 6 on the bottom:

  GND4   A    F       GND3   GND2   B


  E      D  :digit3   C      G      GND1


  2) The reference for the segments:
   AA
  F  B
  F  B
   GG
  E  C
  E  C
   DD


  3) The referencing the digits:
  4  3:  2  1

  So based on this being common cathode, the numbers from 0 to 9, plus clear are:
  0b00111111; // 0; 255-128 = 127
  0b00000110; // 1; =6
  0b01011011; // 2; 2; 64+16+8+2+1 = 91
  0b01001111; // 3; 64+16+4+2+1 = 87
  0b01100110; // 4; 4; 64+32+4+2 = 102
  0b01101101; // 5; 5; 64+32+8+4+1 = 109
  0b01111101; // 6; 255-130 = 125
  0b00000111; // 7; 4+2+1 = 7
  0b01111111; // 8; 8; 127
  0b01100111; // 9; 255-128-16-8 = 103
  0b00000000; // clear 0
  0b10000000; // =128; for the colon by itself, so this number will need to be added to the above

  Then, the digits are the second separate bit to shift out:
  0b00000001; // digit 1 = 1
  0b00000010; // digit 2 = 2
  0b00000100; // digit 3 = 4 (this is the digit with the colon)
  0b00001000; // digit 4 = 8
  These will have to be connected to a transistor (base for NPN BJT)
  which will have the collector connected to the appropriate pin of the display and the emitter connected to GND. The 74HC595 can
  only sink a maximum of 35mA, and the maximum would be '8:' which would involve 8 cathodes supplying current to the ground pin of digit3.

// For the Attiny85
byte clockPin = 0; // pin 11 of shift register
byte latchPin = 1; // pin 12 of the shift register; LOW before shifting out, HIGH when done shifting out
byte dataPin = 2; // pin 14 of shift register
*/
#include <SoftwareSerial.h> // include library
SoftwareSerial tinySerial(3,4); // (rx,tx)
// Where ever this number comes from, it needs 'new line' at the end (does not work with 'carriage return' only)

byte timeValues[] = {0, 0, 0, 0}; //Load the binary value of each digit here
String rawTimeString = "";
int rawTimeInt;
int delayTime = 4; // milliseconds delay before switching to next digit

void getTimeData() {

int tempTime = 0;

while (tinySerial.available() > 0) {
  // Read the incoming data
  char receivedChar = tinySerial.read();

  // Check if the received character is a digit
  if (isdigit(receivedChar)) {
      // Convert the character to an integer and update the received number
      tempTime = tempTime * 10 + (receivedChar - '0');
    } else if (receivedChar == '\n') {
      /* End of the number (newline character received)
      Now save this 3-or-4 digit number to 'rawTimeInt' for further use or processing
      */
      rawTimeInt = tempTime;
      // tinySerial.print("New: "); tinySerial.println(rawTimeInt); tinySerial.println();
    } else {
      // whatever was received was not a number so continue without saving to rawTimeInt      
    }
  }
  }


void parseTimeData(int inputInt) {
  timeValues[3] = (inputInt / 1000);
  int temp = (inputInt % 1000);
  timeValues[2] = (temp / 100);
  temp = (temp % 100);
  timeValues[1] = temp / 10;
  temp = temp % 10;
  timeValues[0] = temp;

// convert each number to a pattern for the display
for (int j=0; j<4; j++) {
  switch (timeValues[j]) {
    case 0: timeValues[j] = 0b00111111; // 0; 255-128 = 127
      break;
    case 1: timeValues[j] = 0b00000110; // 1; =6
      break;
    case 2: timeValues[j] = 0b01011011; // 2; 64+16+8+2+1 = 91
      break;
    case 3: timeValues[j] = 0b01001111; // 3; 64+16+4+2+1 = 87
      break;
    case 4: timeValues[j] = 0b01100110; // 4; 64+32+4+2 = 102
      break;
    case 5: timeValues[j] = 0b01101101; // 5; 64+32+8+4+1 = 109
      break;
    case 6: timeValues[j] = 0b01111101; // 6; 255-130 = 125
      break;
    case 7: timeValues[j] = 0b00000111; // 7; 4+2+1 = 7
      break;
    case 8: timeValues[j] = 0b01111111; // 8; 127
      break;
    case 9: timeValues[j] = 0b01100111; // 9; 255-128-16-8 = 103
      break;
    default: timeValues[j] = 0b00000000; // clear =0
  }
  if (j==2) timeValues[2]=timeValues[2]+0b10000000; // to add the colon
  if (j==3 && (timeValues[3] != 0b00000110)) timeValues[3]=0b00000000; // clears the digit if it isn't a '1'
}
}

byte getDigit(byte number)
{
  byte digit=0b00000000; // without giving an intial value here, an error is generated for 'return digit; below when compilied...

  // remember for 12:59, timeValues[]={1,2,5,9}, so digit1=1 which is the 4th digit from the right of the display
  switch (number) {
    case 1: digit=(0b00000001); // The decimal value representing the binary for digit 1 = 1
      break;
    case 2: digit=(0b00000010); // digit 2 = 2
      break;
    case 3: digit=(0b00000100); // digit 3 = 4 (this is the digit with the colon)
      break;
    case 4: digit=(0b00001000); // digit 4 = 8
      break;
  }
  return digit;
}

void displayTime(byte index)
{
  digitalWrite(latchPin, LOW);
   shiftOut(dataPin, clockPin, MSBFIRST, getDigit(index)); // Shift the digit first as this is downstream of the two 74HC595, these will be 1-4, and will use a npn bjt connected to the shift register
   shiftOut(dataPin, clockPin, MSBFIRST, timeValues[index-1]); // index of timeValues 0-3, so i-1
  digitalWrite(latchPin, HIGH);

   // tinySerial.print(index); tinySerial.print(" "); tinySerial.print(timeValues[index-1]);tinySerial.print(" "); tinySerial.print(timeValues[index-1],BIN); tinySerial.print(" "); tinySerial.print(getDigit(index));tinySerial.print(" "); tinySerial.println(getDigit(index),BIN); 
  /* Reminder regarding shiftOut(datapin,clockpin, bitOrder, value);
   * bitOrder= MSBFIRST or LSBFIRST;
   * value is a byte (not an int), and this can be entered as binary for the pattern of the outputs, or as a number representing that binary pattern:
   * Because the binary number is not being used as a value, but as a pattern of zeros and one to show the pattern.
   * For example, if you only want the first two outputs to be HIGH, then value can be entered as 0b00000011 or as 3 
   * Because 0b00000011 is being used as a pattern with the two LSB as ones, but in binary that pattern also equals the value of 3 (0b00000001 = 1 plus 0b00000010 = 2).
   */
}



void setup() {
  // Here put code required for whatever method is being used to obtain time data: serial monior, RTC, GPS, other microcontroller
tinySerial.begin(9600); // getting time value via serial monitor

  pinMode(latchPin, OUTPUT);
  pinMode(dataPin, OUTPUT);
  pinMode(clockPin, OUTPUT);

}

void loop() {
  
if (tinySerial.available()) {
  // check if new time data received and parse it out
    getTimeData(); // takes serial data from UART, checks for numbers and saves to rawTimeInt which is a 3 or 4 digit number
    // (e.g., raTimeInt of 1234 which will eventually be converted to 12:34)
    
    parseTimeData(rawTimeInt); // takes the rawTimeInt and converts it to an array for each digit of the clock (timeValues[] = {1,2,3,4})
    // tinySerial.println("something received");
  }
  
  for (byte j = 1; j <= 4; j++) {
    displayTime(j); // Will display the time on the display using shiftOut() to the shift registers for the pattern for the digit,
    // using the function getDigit[j] (which is indexed 1-4); then immediately shiftOut() again for timeValues[j-1] (which is indexed 0-3)
    delay(delayTime); // how much delay for each digit one-at-a-time
  }

} // end of loop()

